@@ -51,15 +51,21 @@ typedef struct _PTPContainer PTPContainer;
 
 /* PTP USB Bulk-Pipe container */
 /* USB bulk max packet length for high speed endpoints */
-/* I have changed this from 512 to 4096 now. The spec says
+/* The max packet is set to 512 bytes. The spec says
  * "end of data transfers are signaled by short packets or NULL
- * packets". It never says anything about 512.
- * So it should work like this and the MTP folks will be happier...
- * -Marcus
+ * packets". It never says anything about 512, but current
+ * implementations seem to have chosen this value, which also
+ * happens to be the size of an USB 2.0 HS endpoint, even though
+ * this is not necessary.
+ *
+ * Previously we had this as 4096 for MTP devices. We have found
+ * and fixed the bugs that made this necessary and it can be 512 again.
  */
-#define PTP_USB_BULK_HS_MAX_PACKET_LEN	4096
+#define PTP_USB_BULK_HS_MAX_PACKET_LEN_WRITE	512
+#define PTP_USB_BULK_HS_MAX_PACKET_LEN_READ   512
 #define PTP_USB_BULK_HDR_LEN		(2*sizeof(uint32_t)+2*sizeof(uint16_t))
-#define PTP_USB_BULK_PAYLOAD_LEN	(PTP_USB_BULK_HS_MAX_PACKET_LEN-PTP_USB_BULK_HDR_LEN)
+#define PTP_USB_BULK_PAYLOAD_LEN_WRITE	(PTP_USB_BULK_HS_MAX_PACKET_LEN_WRITE-PTP_USB_BULK_HDR_LEN)
+#define PTP_USB_BULK_PAYLOAD_LEN_READ	(PTP_USB_BULK_HS_MAX_PACKET_LEN_READ-PTP_USB_BULK_HDR_LEN)
 #define PTP_USB_BULK_REQ_LEN	(PTP_USB_BULK_HDR_LEN+5*sizeof(uint32_t))
 
 struct _PTPUSBBulkContainer {
@@ -75,7 +81,9 @@ struct _PTPUSBBulkContainer {
 			uint32_t param4;
 			uint32_t param5;
 		} params;
-		unsigned char data[PTP_USB_BULK_PAYLOAD_LEN];
+       /* this must be set to the maximum of PTP_USB_BULK_PAYLOAD_LEN_WRITE 
+        * and PTP_USB_BULK_PAYLOAD_LEN_READ */
+		unsigned char data[PTP_USB_BULK_PAYLOAD_LEN_READ];
 	} payload;
 };
 typedef struct _PTPUSBBulkContainer PTPUSBBulkContainer;
@@ -92,6 +100,11 @@ struct _PTPUSBEventContainer {
 };
 typedef struct _PTPUSBEventContainer PTPUSBEventContainer;
 
+struct _PTPCanon_directtransfer_entry {
+	uint32_t	oid;
+	char		*str;
+};
+typedef struct _PTPCanon_directtransfer_entry PTPCanon_directtransfer_entry;
 
 /* USB container types */
 
@@ -176,47 +189,63 @@ typedef struct _PTPIPHeader PTPIPHeader;
 #define PTP_OC_EK_SetText		0x9008
 
 /* Canon extension Operation Codes */
-#define PTP_OC_CANON_GetObjectSize	0x9001
-#define PTP_OC_CANON_9002		0x9002
+#define PTP_OC_CANON_GetPartialObjectInfo	0x9001
+#define PTP_OC_CANON_SetObjectArchive		0x9002
 /* 9002 - sends 2 uint32, nothing back  */
-#define PTP_OC_CANON_9003		0x9003
+#define PTP_OC_CANON_9003			0x9003
 /* 9003 - sends nothing, nothing back  */
 /* no 9004 observed yet */
 /* no 9005 observed yet */
-#define PTP_OC_CANON_LookupObject	0x9006
+#define PTP_OC_CANON_GetObjectHandleByName	0x9006
 /* no 9007 observed yet */
-#define PTP_OC_CANON_StartShootingMode	0x9008
-#define PTP_OC_CANON_EndShootingMode	0x9009
+#define PTP_OC_CANON_StartShootingMode		0x9008
+#define PTP_OC_CANON_EndShootingMode		0x9009
 /* 900a - sends nothing, nothing back */
-#define PTP_OC_CANON_900a		0x900A
-#define PTP_OC_CANON_ViewfinderOn	0x900B
-#define PTP_OC_CANON_ViewfinderOff	0x900C
-#define PTP_OC_CANON_ReflectChanges	0x900D
+#define PTP_OC_CANON_900a			0x900A
+#define PTP_OC_CANON_ViewfinderOn		0x900B
+#define PTP_OC_CANON_ViewfinderOff		0x900C
+#define PTP_OC_CANON_DoAeAfAwb			0x900D
 
-/* 900e - send nothing, gets 5 uint16t in 32bit entities back in 20byte datablob 
- * GetFlashDevParam?
- */
-#define PTP_OC_CANON_900e		0x900E
-/* 900f - gets data ... hmm, object handles ? */
-#define PTP_OC_CANON_900f		0x900F
-
-#define PTP_OC_CANON_ThemeDownload	0x9010
-#define PTP_OC_CANON_ThemeUpload	0x9011
+/* 900e - send nothing, gets 5 uint16t in 32bit entities back in 20byte datablob */
+#define PTP_OC_CANON_GetCustomizeSpec		0x900E
+#define PTP_OC_CANON_GetCustomizeItemInfo	0x900F
+#define PTP_OC_CANON_GetCustomizeData		0x9010
+#define PTP_OC_CANON_SetCustomizeData		0x9011
 
 /* initiate movie capture:
    9010 startmoviecapture?
    9003 stopmoviecapture?
 */
 
-#define PTP_OC_CANON_CheckEvent		0x9013
-#define PTP_OC_CANON_FocusLock		0x9014
-#define PTP_OC_CANON_FocusUnlock	0x9015
+#define PTP_OC_CANON_CheckEvent			0x9013
+#define PTP_OC_CANON_FocusLock			0x9014
+#define PTP_OC_CANON_FocusUnlock		0x9015
 #define PTP_OC_CANON_InitiateCaptureInMemory	0x901A
-#define PTP_OC_CANON_GetPartialObject	0x901B
-#define PTP_OC_CANON_GetViewfinderImage	0x901d
-	/* Viewfinder Auto functions */
-#define PTP_OC_CANON_GetChanges		0x9020
-#define PTP_OC_CANON_GetFolderEntries	0x9021
+#define PTP_OC_CANON_GetPartialObjectEx		0x901B
+#define PTP_OC_CANON_SetObjectTime		0x901C
+#define PTP_OC_CANON_GetViewfinderImage		0x901D
+#define PTP_OC_CANON_ChangeUSBProtocol		0x901F
+#define PTP_OC_CANON_GetChanges			0x9020
+#define PTP_OC_CANON_GetObjectInfoEx		0x9021
+
+#define PTP_OC_CANON_TerminateDirectTransfer 	0x9023
+
+#define PTP_OC_CANON_InitiateDirectTransferEx2 	0x9028
+#define PTP_OC_CANON_GetTargetHandles 		0x9029
+#define PTP_OC_CANON_NotifyProgress 		0x902A
+#define PTP_OC_CANON_NotifyCancelAccepted	0x902B
+/* 902c: no parms, read 3 uint32 in data, no response parms */
+#define PTP_OC_CANON_902C			0x902C
+#define PTP_OC_CANON_GetDirectory		0x902D
+
+#define PTP_OC_CANON_SetPairingInfo		0x9030
+#define PTP_OC_CANON_GetPairingInfo		0x9031
+#define PTP_OC_CANON_DeletePairingInfo		0x9032
+#define PTP_OC_CANON_GetMACAddress		0x9033
+/* 9034: 1 param, no parms returned */
+#define PTP_OC_CANON_SetDisplayMonitor		0x9034
+#define PTP_OC_CANON_PairingComplete		0x9035
+#define PTP_OC_CANON_GetWirelessMAXChannel	0x9036
 
 /* Nikon extension Operation Codes */
 #define PTP_OC_NIKON_GetProfileAllData	0x9006
@@ -234,6 +263,7 @@ typedef struct _PTPIPHeader PTPIPHeader;
 #define PTP_OC_NIKON_GetDevicePTPIPInfo	0x90E0
 
 /* Microsoft / MTP extension codes */
+
 #define PTP_OC_MTP_GetObjectPropsSupported	0x9801
 #define PTP_OC_MTP_GetObjectPropDesc		0x9802
 #define PTP_OC_MTP_GetObjectPropValue		0x9803
@@ -245,7 +275,73 @@ typedef struct _PTPIPHeader PTPIPHeader;
 #define PTP_OC_MTP_GetObjectReferences		0x9810
 #define PTP_OC_MTP_SetObjectReferences		0x9811
 #define PTP_OC_MTP_UpdateDeviceFirmware		0x9812
-#define PTP_OC_MTP_Skip                         0x9820
+#define PTP_OC_MTP_Skip				0x9820
+
+/*
+ * Windows Media Digital Rights Management for Portable Devices 
+ * Extension Codes (microsoft.com/WMDRMPD: 10.1)
+ */
+#define PTP_OC_MTP_WMDRMPD_GetSecureTimeChallenge	0x9101
+#define PTP_OC_MTP_WMDRMPD_GetSecureTimeResponse	0x9102
+#define PTP_OC_MTP_WMDRMPD_SetLicenseResponse	0x9103
+#define PTP_OC_MTP_WMDRMPD_GetSyncList		0x9104
+#define PTP_OC_MTP_WMDRMPD_SendMeterChallengeQuery	0x9105
+#define PTP_OC_MTP_WMDRMPD_GetMeterChallenge	0x9106
+#define PTP_OC_MTP_WMDRMPD_SetMeterResponse		0x9107
+#define PTP_OC_MTP_WMDRMPD_CleanDataStore		0x9108
+#define PTP_OC_MTP_WMDRMPD_GetLicenseState		0x9109
+#define PTP_OC_MTP_WMDRMPD_SendWMDRMPDCommand	0x910A
+#define PTP_OC_MTP_WMDRMPD_SendWMDRMPDRequest	0x910B
+
+/* 
+ * Windows Media Digital Rights Management for Portable Devices 
+ * Extension Codes (microsoft.com/WMDRMPD: 10.1)
+ * Below are operations that have no public documented identifier 
+ * associated with them "Vendor-defined Command Code"
+ */
+#define PTP_OC_MTP_WMDRMPD_SendWMDRMPDAppRequest	0x9212
+#define PTP_OC_MTP_WMDRMPD_GetWMDRMPDAppResponse	0x9213
+#define PTP_OC_MTP_WMDRMPD_EnableTrustedFilesOperations	0x9214
+#define PTP_OC_MTP_WMDRMPD_DisableTrustedFilesOperations 0x9215
+#define PTP_OC_MTP_WMDRMPD_EndTrustedAppSession		0x9216
+/* ^^^ guess ^^^ */
+
+/*
+ * Microsoft Advanced Audio/Video Transfer 
+ * Extensions (microsoft.com/AAVT: 1.0)
+ */
+#define PTP_OC_MTP_AAVT_OpenMediaSession		0x9170
+#define PTP_OC_MTP_AAVT_CloseMediaSession		0x9171
+#define PTP_OC_MTP_AAVT_GetNextDataBlock		0x9172
+#define PTP_OC_MTP_AAVT_SetCurrentTimePosition		0x9173
+
+/*
+ * Windows Media Digital Rights Management for Network Devices 
+ * Extensions (microsoft.com/WMDRMND: 1.0) MTP/IP?
+ */
+#define PTP_OC_MTP_WMDRMND_SendRegistrationRequest	0x9180
+#define PTP_OC_MTP_WMDRMND_GetRegistrationResponse	0x9181
+#define PTP_OC_MTP_WMDRMND_GetProximityChallenge	0x9182
+#define PTP_OC_MTP_WMDRMND_SendProximityResponse	0x9183
+#define PTP_OC_MTP_WMDRMND_SendWMDRMNDLicenseRequest	0x9184
+#define PTP_OC_MTP_WMDRMND_GetWMDRMNDLicenseResponse	0x9185
+
+/* 
+ * Windows Media Player Portiable Devices 
+ * Extension Codes (microsoft.com/WMPPD: 11.1)
+ */
+#define PTP_OC_MTP_WMPPD_ReportAddedDeletedItems	0x9201
+#define PTP_OC_MTP_WMPPD_ReportAcquiredItems 	        0x9202
+#define PTP_OC_MTP_WMPPD_PlaylistObjectPref		0x9203
+
+/*
+ * Undocumented Zune Operation Codes 
+ * maybe related to WMPPD extension set?
+ */
+#define PTP_OC_MTP_ZUNE_GETUNDEFINED001		        0x9204
+
+/* WiFi Provisioning MTP Extension Codes (microsoft.com/WPDWCN: 1.0) */
+#define PTP_OC_MTP_WPDWCN_ProcessWFCObject		0x9122
 
 /* Proprietary vendor extension operations mask */
 #define PTP_OC_EXTENSION_MASK           0xF000
@@ -305,6 +401,16 @@ typedef struct _PTPIPHeader PTPIPHeader;
 #define PTP_RC_MTP_Specification_By_Group_Unsupported		0xA808
 #define PTP_RC_MTP_Object_Too_Large		0xA809
 
+/* Microsoft Advanced Audio/Video Transfer response codes 
+(microsoft.com/AAVT 1.0) */
+#define PTP_RC_MTP_Invalid_Media_Session_ID	0xA170	
+#define PTP_RC_MTP_Media_Session_Limit_Reached	0xA171
+#define PTP_RC_MTP_No_More_Data			0xA172
+
+/* WiFi Provisioning MTP Extension Error Codes (microsoft.com/WPDWCN: 1.0) */
+#define PTP_RC_MTP_Invalid_WFC_Syntax		0xA121
+#define PTP_RC_MTP_WFC_Version_Not_Supported	0xA122
+
 /* libptp2 extended ERROR codes */
 #define PTP_ERROR_IO			0x02FF
 #define PTP_ERROR_DATA_EXPECTED		0x02FE
@@ -335,8 +441,12 @@ typedef struct _PTPIPHeader PTPIPHeader;
 #define PTP_EC_CANON_RequestObjectTransfer	0xC009
 #define PTP_EC_CANON_CameraModeChanged		0xC00C
 
+#define PTP_EC_CANON_StartDirectTransfer	0xC011
+#define PTP_EC_CANON_StopDirectTransfer		0xC013
+
 /* Nikon extension Event Codes */
 #define PTP_EC_Nikon_ObjectAddedInSDRAM		0xC101
+#define PTP_EC_Nikon_CaptureOverflow		0xC102
 /* Gets 1 parameter, objectid pointing to DPOF object */
 #define PTP_EC_Nikon_AdvancedTransfer		0xC103
 
@@ -348,7 +458,7 @@ typedef struct _PTPIPHeader PTPIPHeader;
 /* PTP device info structure (returned by GetDevInfo) */
 
 struct _PTPDeviceInfo {
-	uint16_t StaqndardVersion;
+	uint16_t StandardVersion;
 	uint32_t VendorExtensionID;
 	uint16_t VendorExtensionVersion;
 	char	*VendorExtensionDesc;
@@ -436,6 +546,7 @@ typedef struct _PTPObjectInfo PTPObjectInfo;
 
 /* ancillary formats */
 #define PTP_OFC_Undefined			0x3000
+#define PTP_OFC_Defined				0x3800
 #define PTP_OFC_Association			0x3001
 #define PTP_OFC_Script				0x3002
 #define PTP_OFC_Executable			0x3003
@@ -469,15 +580,25 @@ typedef struct _PTPObjectInfo PTPObjectInfo;
 /* Eastman Kodak extension ancillary format */
 #define PTP_OFC_EK_M3U				0xb002
 /* MTP extensions */
+#define PTP_OFC_MTP_MediaCard			0xb211
+#define PTP_OFC_MTP_MediaCardGroup		0xb212
+#define PTP_OFC_MTP_Encounter			0xb213
+#define PTP_OFC_MTP_EncounterBox		0xb214
+#define PTP_OFC_MTP_M4A				0xb215
+#define PTP_OFC_MTP_ZUNEUNDEFINED		0xb217 /* Unknown file type */
 #define PTP_OFC_MTP_Firmware			0xb802
-#define PTP_OFC_MTP_WindowsImageFormat		0xb882
+#define PTP_OFC_MTP_WindowsImageFormat		0xb881
 #define PTP_OFC_MTP_UndefinedAudio		0xb900
 #define PTP_OFC_MTP_WMA				0xb901
 #define PTP_OFC_MTP_OGG				0xb902
+#define PTP_OFC_MTP_AAC				0xb903
 #define PTP_OFC_MTP_AudibleCodec		0xb904
+#define PTP_OFC_MTP_FLAC			0xb906
 #define PTP_OFC_MTP_UndefinedVideo		0xb980
 #define PTP_OFC_MTP_WMV				0xb981
 #define PTP_OFC_MTP_MP4				0xb982
+#define PTP_OFC_MTP_MP2				0xb983
+#define PTP_OFC_MTP_3GP				0xb984
 #define PTP_OFC_MTP_UndefinedCollection		0xba00
 #define PTP_OFC_MTP_AbstractMultimediaAlbum	0xba01
 #define PTP_OFC_MTP_AbstractImageAlbum		0xba02
@@ -487,6 +608,9 @@ typedef struct _PTPObjectInfo PTPObjectInfo;
 #define PTP_OFC_MTP_AbstractContactGroup	0xba06
 #define PTP_OFC_MTP_AbstractMessageFolder	0xba07
 #define PTP_OFC_MTP_AbstractChapteredProduction	0xba08
+#define PTP_OFC_MTP_AbstractAudioPlaylist	0xba09
+#define PTP_OFC_MTP_AbstractVideoPlaylist	0xba0a
+#define PTP_OFC_MTP_AbstractMediacast		0xba0b
 #define PTP_OFC_MTP_WPLPlaylist			0xba10
 #define PTP_OFC_MTP_M3UPlaylist			0xba11
 #define PTP_OFC_MTP_MPLPlaylist			0xba12
@@ -494,6 +618,11 @@ typedef struct _PTPObjectInfo PTPObjectInfo;
 #define PTP_OFC_MTP_PLSPlaylist			0xba14
 #define PTP_OFC_MTP_UndefinedDocument		0xba80
 #define PTP_OFC_MTP_AbstractDocument		0xba81
+#define PTP_OFC_MTP_XMLDocument			0xba82
+#define PTP_OFC_MTP_MSWordDocument		0xba83
+#define PTP_OFC_MTP_MHTCompiledHTMLDocument	0xba84
+#define PTP_OFC_MTP_MSExcelSpreadsheetXLS	0xba85
+#define PTP_OFC_MTP_MSPowerpointPresentationPPT	0xba86
 #define PTP_OFC_MTP_UndefinedMessage		0xbb00
 #define PTP_OFC_MTP_AbstractMessage		0xbb01
 #define PTP_OFC_MTP_UndefinedContact		0xbb80
@@ -506,6 +635,7 @@ typedef struct _PTPObjectInfo PTPObjectInfo;
 #define PTP_OFC_MTP_vCalendar2			0xbe03
 #define PTP_OFC_MTP_UndefinedWindowsExecutable	0xbe80
 
+
 /* PTP Association Types */
 #define PTP_AT_Undefined			0x0000
 #define PTP_AT_GenericFolder			0x0001
@@ -546,7 +676,6 @@ typedef struct _PTPObjectInfo PTPObjectInfo;
 
 union _PTPPropertyValue {
 	char		*str;	/* common string, malloced */
-	uint16_t        *unistr; /* UCS2 unicode string, malloced */
 	uint8_t		u8;
 	int8_t		i8;
 	uint16_t	u16;
@@ -562,6 +691,16 @@ union _PTPPropertyValue {
 
 typedef union _PTPPropertyValue PTPPropertyValue;
 
+/* Metadata lists for MTP operations */
+struct _MTPPropList {
+	uint16_t 	 	property;
+	uint16_t 	 	datatype;
+	uint32_t 	 	ObjectHandle;
+	PTPPropertyValue 	propval;
+	struct _MTPPropList 	*next;
+};
+typedef struct _MTPPropList MTPPropList;
+
 struct _PTPPropDescRangeForm {
 	PTPPropertyValue 	MinimumValue;
 	PTPPropertyValue 	MaximumValue;
@@ -763,7 +902,11 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_DPC_EXTENSION_MASK		0xF000
 #define PTP_DPC_EXTENSION		0xD000
 
-/* Vendor Extensions device property codes */
+/* Zune extension device property codes */
+#define PTP_DPC_MTP_ZUNE_UNKNOWN1	0xD181
+#define PTP_DPC_MTP_ZUNE_UNKNOWN2	0xD132
+#define PTP_DPC_MTP_ZUNE_UNKNOWN3	0xD215
+#define PTP_DPC_MTP_ZUNE_UNKNOWN4	0xD216
 
 /* Eastman Kodak extension device property codes */
 #define PTP_DPC_EK_ColorTemperature	0xD001
@@ -866,6 +1009,7 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_DPC_NIKON_E6ManualModeBracketing		0xD079	
 #define PTP_DPC_NIKON_BracketOrder			0xD07A
 #define PTP_DPC_NIKON_E8AutoBracketSelection		0xD07B
+#define PTP_DPC_NIKON_BracketingSet			0xD07C
 #define PTP_DPC_NIKON_F1CenterButtonShootingMode	0xD080
 #define PTP_DPC_NIKON_CenterButtonPlaybackMode		0xD081
 #define PTP_DPC_NIKON_F2Multiselector			0xD082
@@ -883,13 +1027,16 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_DPC_NIKON_Bracketing			0xD0C0
 #define PTP_DPC_NIKON_ExposureBracketingIntervalDist	0xD0C1
 #define PTP_DPC_NIKON_BracketingProgram			0xD0C2
+#define PTP_DPC_NIKON_WhiteBalanceBracketStep		0xD0C4
 #define PTP_DPC_NIKON_LensID				0xD0E0
 #define PTP_DPC_NIKON_FocalLengthMin			0xD0E3
 #define PTP_DPC_NIKON_FocalLengthMax			0xD0E4
 #define PTP_DPC_NIKON_MaxApAtMinFocalLength		0xD0E5
 #define PTP_DPC_NIKON_MaxApAtMaxFocalLength		0xD0E6
 #define PTP_DPC_NIKON_ExposureTime			0xD100
+#define PTP_DPC_NIKON_ACPower				0xD101
 #define PTP_DPC_NIKON_MaximumShots			0xD103
+#define PTP_DPC_NIKON_AFLLock				0xD104
 #define PTP_DPC_NIKON_AutoExposureLock			0xD105
 #define PTP_DPC_NIKON_AutoFocusLock			0xD106
 #define PTP_DPC_NIKON_AutofocusLCDTopMode2		0xD107
@@ -922,6 +1069,7 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 /* Microsoft/MTP specific */
 #define PTP_DPC_MTP_SecureTime                          0xD101
 #define PTP_DPC_MTP_DeviceCertificate                   0xD102
+#define PTP_DPC_MTP_RevocationInfo                      0xD103
 #define PTP_DPC_MTP_SynchronizationPartner              0xD401
 #define PTP_DPC_MTP_DeviceFriendlyName                  0xD402
 #define PTP_DPC_MTP_VolumeLevel                         0xD403
@@ -930,6 +1078,10 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_DPC_MTP_PlaybackObject                      0xD411
 #define PTP_DPC_MTP_PlaybackContainerIndex              0xD412
 #define PTP_DPC_MTP_PlaybackPosition                    0xD413
+#define PTP_DPC_MTP_PlaysForSureID                      0xD131
+
+/* Zune specific property codes */
+#define PTP_DPC_MTP_Zune_UnknownVersion			0xD181
 
 /* MTP specific Object Properties */
 #define PTP_OPC_StorageID				0xDC01
@@ -943,6 +1095,9 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_OPC_DateModified				0xDC09
 #define PTP_OPC_Keywords				0xDC0A
 #define PTP_OPC_ParentObject				0xDC0B
+#define PTP_OPC_AllowedFolderContents			0xDC0C
+#define PTP_OPC_Hidden					0xDC0D
+#define PTP_OPC_SystemObject				0xDC0E
 #define PTP_OPC_PersistantUniqueObjectIdentifier	0xDC41
 #define PTP_OPC_SyncID					0xDC42
 #define PTP_OPC_PropertyBag				0xDC43
@@ -959,6 +1114,7 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_OPC_DateAdded				0xDC4E
 #define PTP_OPC_NonConsumable				0xDC4F
 #define PTP_OPC_CorruptOrUnplayable			0xDC50
+#define PTP_OPC_ProducerSerialNumber			0xDC51
 #define PTP_OPC_RepresentativeSampleFormat		0xDC81
 #define PTP_OPC_RepresentativeSampleSize		0xDC82
 #define PTP_OPC_RepresentativeSampleHeight		0xDC83
@@ -991,6 +1147,94 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_OPC_SubDescription				0xDC9E
 #define PTP_OPC_IsCropped				0xDCD1
 #define PTP_OPC_IsColorCorrected			0xDCD2
+#define PTP_OPC_ImageBitDepth				0xDCD3
+#define PTP_OPC_Fnumber					0xDCD4
+#define PTP_OPC_ExposureTime				0xDCD5
+#define PTP_OPC_ExposureIndex				0xDCD6
+#define PTP_OPC_DisplayName				0xDCE0
+#define PTP_OPC_BodyText				0xDCE1
+#define PTP_OPC_Subject					0xDCE2
+#define PTP_OPC_Prority					0xDCE3
+#define PTP_OPC_GivenName				0xDD00
+#define PTP_OPC_MiddleNames				0xDD01
+#define PTP_OPC_FamilyName				0xDD02
+#define PTP_OPC_Prefix					0xDD03
+#define PTP_OPC_Suffix					0xDD04
+#define PTP_OPC_PhoneticGivenName			0xDD05
+#define PTP_OPC_PhoneticFamilyName			0xDD06
+#define PTP_OPC_EmailPrimary				0xDD07
+#define PTP_OPC_EmailPersonal1				0xDD08
+#define PTP_OPC_EmailPersonal2				0xDD09
+#define PTP_OPC_EmailBusiness1				0xDD0A
+#define PTP_OPC_EmailBusiness2				0xDD0B
+#define PTP_OPC_EmailOthers				0xDD0C
+#define PTP_OPC_PhoneNumberPrimary			0xDD0D
+#define PTP_OPC_PhoneNumberPersonal			0xDD0E
+#define PTP_OPC_PhoneNumberPersonal2			0xDD0F
+#define PTP_OPC_PhoneNumberBusiness			0xDD10
+#define PTP_OPC_PhoneNumberBusiness2			0xDD11
+#define PTP_OPC_PhoneNumberMobile			0xDD12
+#define PTP_OPC_PhoneNumberMobile2			0xDD13
+#define PTP_OPC_FaxNumberPrimary			0xDD14
+#define PTP_OPC_FaxNumberPersonal			0xDD15
+#define PTP_OPC_FaxNumberBusiness			0xDD16
+#define PTP_OPC_PagerNumber				0xDD17
+#define PTP_OPC_PhoneNumberOthers			0xDD18
+#define PTP_OPC_PrimaryWebAddress			0xDD19
+#define PTP_OPC_PersonalWebAddress			0xDD1A
+#define PTP_OPC_BusinessWebAddress			0xDD1B
+#define PTP_OPC_InstantMessengerAddress			0xDD1C
+#define PTP_OPC_InstantMessengerAddress2		0xDD1D
+#define PTP_OPC_InstantMessengerAddress3		0xDD1E
+#define PTP_OPC_PostalAddressPersonalFull		0xDD1F
+#define PTP_OPC_PostalAddressPersonalFullLine1		0xDD20
+#define PTP_OPC_PostalAddressPersonalFullLine2		0xDD21
+#define PTP_OPC_PostalAddressPersonalFullCity		0xDD22
+#define PTP_OPC_PostalAddressPersonalFullRegion		0xDD23
+#define PTP_OPC_PostalAddressPersonalFullPostalCode	0xDD24
+#define PTP_OPC_PostalAddressPersonalFullCountry	0xDD25
+#define PTP_OPC_PostalAddressBusinessFull		0xDD26
+#define PTP_OPC_PostalAddressBusinessLine1		0xDD27
+#define PTP_OPC_PostalAddressBusinessLine2		0xDD28
+#define PTP_OPC_PostalAddressBusinessCity		0xDD29
+#define PTP_OPC_PostalAddressBusinessRegion		0xDD2A
+#define PTP_OPC_PostalAddressBusinessPostalCode		0xDD2B
+#define PTP_OPC_PostalAddressBusinessCountry		0xDD2C
+#define PTP_OPC_PostalAddressOtherFull			0xDD2D
+#define PTP_OPC_PostalAddressOtherLine1			0xDD2E
+#define PTP_OPC_PostalAddressOtherLine2			0xDD2F
+#define PTP_OPC_PostalAddressOtherCity			0xDD30
+#define PTP_OPC_PostalAddressOtherRegion		0xDD31
+#define PTP_OPC_PostalAddressOtherPostalCode		0xDD32
+#define PTP_OPC_PostalAddressOtherCountry		0xDD33
+#define PTP_OPC_OrganizationName			0xDD34
+#define PTP_OPC_PhoneticOrganizationName		0xDD35
+#define PTP_OPC_Role					0xDD36
+#define PTP_OPC_Birthdate				0xDD37
+#define PTP_OPC_MessageTo				0xDD40
+#define PTP_OPC_MessageCC				0xDD41
+#define PTP_OPC_MessageBCC				0xDD42
+#define PTP_OPC_MessageRead				0xDD43
+#define PTP_OPC_MessageReceivedTime			0xDD44
+#define PTP_OPC_MessageSender				0xDD45
+#define PTP_OPC_ActivityBeginTime			0xDD50
+#define PTP_OPC_ActivityEndTime				0xDD51
+#define PTP_OPC_ActivityLocation			0xDD52
+#define PTP_OPC_ActivityRequiredAttendees		0xDD54
+#define PTP_OPC_ActivityOptionalAttendees		0xDD55
+#define PTP_OPC_ActivityResources			0xDD56
+#define PTP_OPC_ActivityAccepted			0xDD57
+#define PTP_OPC_Owner					0xDD5D
+#define PTP_OPC_Editor					0xDD5E
+#define PTP_OPC_Webmaster				0xDD5F
+#define PTP_OPC_URLSource				0xDD60
+#define PTP_OPC_URLDestination				0xDD61
+#define PTP_OPC_TimeBookmark				0xDD62
+#define PTP_OPC_ObjectBookmark				0xDD63
+#define PTP_OPC_ByteBookmark				0xDD64
+#define PTP_OPC_LastBuildDate				0xDD70
+#define PTP_OPC_TimetoLive				0xDD71
+#define PTP_OPC_MediaGUID				0xDD72
 #define PTP_OPC_TotalBitRate				0xDE91
 #define PTP_OPC_BitRateType				0xDE92
 #define PTP_OPC_SampleRate				0xDE93
@@ -1005,6 +1249,11 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 #define PTP_OPC_KeyFrameDistance			0xDE9E
 #define PTP_OPC_BufferSize				0xDE9F
 #define PTP_OPC_EncodingQuality				0xDEA0
+#define PTP_OPC_EncodingProfile				0xDEA1
+#define PTP_OPC_BuyFlag					0xD901
+
+/* WiFi Provisioning MTP Extension property codes */
+#define PTP_OPC_WirelessConfigurationFile		0xB104
 
 /* Device Property Form Flag */
 
@@ -1030,23 +1279,31 @@ typedef struct _PTPNIKONWifiProfile PTPNIKONWifiProfile;
 
 typedef struct _PTPParams PTPParams;
 
-/* raw write functions */
-typedef short (* PTPIOReadFunc)	(unsigned char *bytes, unsigned int size,
-				 void *data, unsigned int *readlen);
-typedef short (* PTPIOWriteFunc)(unsigned char *bytes, unsigned int size,
-				 void *data);
+
+typedef uint16_t (* PTPDataGetFunc)	(PTPParams* params, void* private,
+					unsigned long wantlen,
+	                                unsigned char *data, unsigned long *gotlen);
+
+typedef uint16_t (* PTPDataPutFunc)	(PTPParams* params, void* private,
+					unsigned long sendlen,
+	                                unsigned char *data, unsigned long *putlen);
+typedef struct _PTPDataHandler {
+	PTPDataGetFunc		getfunc;
+	PTPDataPutFunc		putfunc;
+	void			*private;
+} PTPDataHandler;
+
 /*
  * This functions take PTP oriented arguments and send them over an
  * appropriate data layer doing byteorder conversion accordingly.
  */
 typedef uint16_t (* PTPIOSendReq)	(PTPParams* params, PTPContainer* req);
 typedef uint16_t (* PTPIOSendData)	(PTPParams* params, PTPContainer* ptp,
-					 unsigned char *data, unsigned int size,
-					 int from_fd);
+					 unsigned long size, PTPDataHandler*getter);
+
 typedef uint16_t (* PTPIOGetResp)	(PTPParams* params, PTPContainer* resp);
 typedef uint16_t (* PTPIOGetData)	(PTPParams* params, PTPContainer* ptp,
-	                                unsigned char **data, unsigned int *recvlen,
-	                                int to_fd);
+	                                 PTPDataHandler *putter);
 /* debug functions */
 typedef void (* PTPErrorFunc) (void *data, const char *format, va_list args)
 #if (__GNUC__ >= 3)
@@ -1061,15 +1318,10 @@ typedef void (* PTPDebugFunc) (void *data, const char *format, va_list args)
 
 struct _PTPParams {
 	/* data layer byteorder */
-	uint8_t	byteorder;
+	uint8_t		byteorder;
+	uint16_t	maxpacketsize;
 
-	/* Data layer IO functions */
-	PTPIOReadFunc	read_func;
-	PTPIOWriteFunc	write_func;
-	PTPIOReadFunc	check_int_func;
-	PTPIOReadFunc	check_int_fast_func;
-
-	/* Custom IO functions */
+	/* PTP IO: Custom IO functions */
 	PTPIOSendReq	sendreq_func;
 	PTPIOSendData	senddata_func;
 	PTPIOGetResp	getresp_func;
@@ -1078,62 +1330,70 @@ struct _PTPParams {
 	PTPIOGetResp	event_wait;
 
 	/* Custom error and debug function */
-	PTPErrorFunc error_func;
-	PTPDebugFunc debug_func;
+	PTPErrorFunc	error_func;
+	PTPDebugFunc	debug_func;
 
 	/* Data passed to above functions */
-	void *data;
+	void		*data;
 
 	/* ptp transaction ID */
-	uint32_t transaction_id;
+	uint32_t	transaction_id;
 	/* ptp session ID */
-	uint32_t session_id;
+	uint32_t	session_id;
 
-	/* if we have MTP style split header/data transfers */
-	int	split_header_data;
+	/* PTP IO: if we have MTP style split header/data transfers */
+	int		split_header_data;
 
-	/* internal structures used by ptp driver */
+	/* PTP: MTP specific structure. */
+	MTPPropList	*proplist;
+
+	/* PTP: internal structures used by ptp driver */
 	PTPObjectHandles handles;
-	PTPObjectInfo * objectinfo;
-	PTPDeviceInfo deviceinfo;
+	PTPObjectInfo	*objectinfo;
+	PTPDeviceInfo	deviceinfo;
 
-	/* Wifi profiles */
-	uint8_t  wifi_profiles_version;
-	uint8_t  wifi_profiles_number;
+	/* PTP: Canon specific flags list */
+	uint32_t	*canon_flags; /* size(handles.n) */
+
+	/* PTP: Wifi profiles */
+	uint8_t 	wifi_profiles_version;
+	uint8_t		wifi_profiles_number;
 	PTPNIKONWifiProfile *wifi_profiles;
 
-	/* PTP/IP related data */
+	/* IO: PTP/IP related data */
 	int		cmdfd, evtfd;
 	uint8_t		cameraguid[16];
 	uint32_t	eventpipeid;
 	char		*cameraname;
 
-	/* iconv converters */
-	iconv_t cd_locale_to_ucs2;
+	/* PTP: iconv converters */
+	iconv_t	cd_locale_to_ucs2;
 	iconv_t cd_ucs2_to_locale;
+
+	/* IO: Sometimes the response packet get send in the dataphase
+	 * too. This only happens for a Samsung player now.
+	 */
+	uint8_t		*response_packet;
+	uint16_t	response_packet_size;
 };
 
 /* last, but not least - ptp functions */
 uint16_t ptp_usb_sendreq	(PTPParams* params, PTPContainer* req);
 uint16_t ptp_usb_senddata	(PTPParams* params, PTPContainer* ptp,
-				 unsigned char *data, unsigned int size,
-				 int from_fd);
+				 unsigned long size, PTPDataHandler *handler);
 uint16_t ptp_usb_getresp	(PTPParams* params, PTPContainer* resp);
 uint16_t ptp_usb_getdata	(PTPParams* params, PTPContainer* ptp, 
-				 unsigned char **data, unsigned int *readlen,
-	                         int to_fd);
+	                         PTPDataHandler *handler);
 uint16_t ptp_usb_event_check	(PTPParams* params, PTPContainer* event);
 uint16_t ptp_usb_event_wait	(PTPParams* params, PTPContainer* event);
 
 int      ptp_ptpip_connect	(PTPParams* params, const char *port);
 uint16_t ptp_ptpip_sendreq	(PTPParams* params, PTPContainer* req);
 uint16_t ptp_ptpip_senddata	(PTPParams* params, PTPContainer* ptp,
-				unsigned char *data, unsigned int size,
-				int to_fd);
+				unsigned long size, PTPDataHandler *handler);
 uint16_t ptp_ptpip_getresp	(PTPParams* params, PTPContainer* resp);
 uint16_t ptp_ptpip_getdata	(PTPParams* params, PTPContainer* ptp, 
-				 unsigned char **data, unsigned int *readlen,
-	                         int to_fd);
+	                         PTPDataHandler *handler);
 uint16_t ptp_ptpip_event_wait	(PTPParams* params, PTPContainer* event);
 uint16_t ptp_ptpip_event_check	(PTPParams* params, PTPContainer* event);
 
@@ -1163,6 +1423,7 @@ uint16_t ptp_getobjectinfo	(PTPParams *params, uint32_t handle,
 uint16_t ptp_getobject		(PTPParams *params, uint32_t handle,
 				unsigned char** object);
 uint16_t ptp_getobject_tofd     (PTPParams* params, uint32_t handle, int fd);
+uint16_t ptp_getobject_to_handler (PTPParams* params, uint32_t handle, PTPDataHandler*);
 uint16_t ptp_getpartialobject	(PTPParams* params, uint32_t handle, uint32_t offset,
 				uint32_t maxbytes, unsigned char** object);
 uint16_t ptp_getthumb		(PTPParams *params, uint32_t handle,
@@ -1177,6 +1438,7 @@ uint16_t ptp_sendobjectinfo	(PTPParams* params, uint32_t* store,
 uint16_t ptp_sendobject		(PTPParams* params, unsigned char* object,
 				 uint32_t size);
 uint16_t ptp_sendobject_fromfd  (PTPParams* params, int fd, uint32_t size);
+uint16_t ptp_sendobject_from_handler  (PTPParams* params, PTPDataHandler*, uint32_t size);
 
 uint16_t ptp_initiatecapture	(PTPParams* params, uint32_t storageid,
 				uint32_t ofc);
@@ -1188,6 +1450,7 @@ uint16_t ptp_getdevicepropvalue	(PTPParams* params, uint16_t propcode,
 uint16_t ptp_setdevicepropvalue (PTPParams* params, uint16_t propcode,
                         	PTPPropertyValue* value, uint16_t datatype);
 
+/* Microsoft MTP extensions */
 uint16_t ptp_mtp_getobjectpropdesc (PTPParams* params, uint16_t opc, uint16_t ofc,
 				PTPObjectPropDesc *objectpropertydesc);
 uint16_t ptp_mtp_getobjectpropvalue (PTPParams* params, uint32_t oid, uint16_t opc, 
@@ -1196,7 +1459,9 @@ uint16_t ptp_mtp_setobjectpropvalue (PTPParams* params, uint32_t oid, uint16_t o
 				PTPPropertyValue *value, uint16_t datatype);
 uint16_t ptp_mtp_getobjectreferences (PTPParams* params, uint32_t handle, uint32_t** ohArray, uint32_t* arraylen);
 uint16_t ptp_mtp_setobjectreferences (PTPParams* params, uint32_t handle, uint32_t* ohArray, uint32_t arraylen);
-
+uint16_t ptp_mtp_getobjectproplist (PTPParams* params, uint32_t handle, MTPPropList **proplist);
+uint16_t ptp_mtp_sendobjectproplist (PTPParams* params, uint32_t* store, uint32_t* parenthandle, uint32_t* handle,
+				     uint16_t objecttype, uint64_t objectsize, MTPPropList *proplist);
 
 /* Eastman Kodak extensions */
 uint16_t ptp_ek_9007 (PTPParams* params, unsigned char **serial, unsigned int *size);
@@ -1210,19 +1475,24 @@ uint16_t ptp_ek_sendfileobjectinfo (PTPParams* params, uint32_t* store,
 				PTPObjectInfo* objectinfo);
 uint16_t ptp_ek_sendfileobject	(PTPParams* params, unsigned char* object,
 				uint32_t size);
+uint16_t ptp_ek_sendfileobject_from_handler	(PTPParams* params, PTPDataHandler*,
+				uint32_t size);
 
 /* Canon PTP extensions */
 uint16_t ptp_canon_9012 (PTPParams* params);
-uint16_t ptp_canon_getobjectsize (PTPParams* params, uint32_t handle,
+uint16_t ptp_canon_initiate_direct_transfer (PTPParams* params, uint32_t* out);
+uint16_t ptp_canon_get_target_handles (PTPParams* params, PTPCanon_directtransfer_entry**, unsigned int*cnt);
+uint16_t ptp_canon_getpartialobjectinfo (PTPParams* params, uint32_t handle,
 				uint32_t p2, uint32_t* size, uint32_t* rp2);
 
+uint16_t ptp_canon_get_mac_address (PTPParams* params, unsigned char **mac);
 uint16_t ptp_canon_startshootingmode (PTPParams* params);
 uint16_t ptp_canon_endshootingmode (PTPParams* params);
 
 uint16_t ptp_canon_viewfinderon (PTPParams* params);
 uint16_t ptp_canon_viewfinderoff (PTPParams* params);
 
-uint16_t ptp_canon_reflectchanges (PTPParams* params, uint32_t p1);
+uint16_t ptp_canon_aeafawb (PTPParams* params, uint32_t p1);
 uint16_t ptp_canon_checkevent (PTPParams* params, 
 				PTPUSBEventContainer* event, int* isevent);
 uint16_t ptp_canon_focuslock (PTPParams* params);
@@ -1236,15 +1506,18 @@ uint16_t ptp_canon_getviewfinderimage (PTPParams* params, unsigned char** image,
 				uint32_t* size);
 uint16_t ptp_canon_getchanges (PTPParams* params, uint16_t** props,
 				uint32_t* propnum); 
-uint16_t ptp_canon_getfolderentries (PTPParams* params, uint32_t store,
+uint16_t ptp_canon_getobjectinfo (PTPParams* params, uint32_t store,
 				uint32_t p2, uint32_t parenthandle,
 				uint32_t handle, 
 				PTPCANONFolderEntry** entries,
 				uint32_t* entnum);
-uint16_t ptp_canon_lookup_object (PTPParams* params, char* name, uint32_t* objectid);
-uint16_t ptp_canon_theme_download (PTPParams* params, uint32_t themenr,
+uint16_t ptp_canon_get_objecthandle_by_name (PTPParams* params, char* name, uint32_t* objectid);
+uint16_t ptp_canon_get_directory (PTPParams* params, PTPObjectHandles *handles, PTPObjectInfo **oinfos, uint32_t **flags);
+uint16_t ptp_canon_setobjectarchive (PTPParams* params, uint32_t oid, uint32_t flags);
+uint16_t ptp_canon_get_customize_data (PTPParams* params, uint32_t themenr,
 				unsigned char **data, unsigned int *size);
 
+
 uint16_t ptp_nikon_curve_download (PTPParams* params, 
 				unsigned char **data, unsigned int *size);
 uint16_t ptp_nikon_getptpipinfo (PTPParams* params, unsigned char **data, unsigned int *size);
@@ -1267,6 +1540,8 @@ int ptp_property_issupported	(PTPParams* params, uint16_t property);
 void ptp_free_devicepropdesc	(PTPDevicePropDesc* dpd);
 void ptp_free_devicepropvalue	(uint16_t dt, PTPPropertyValue* dpd);
 void ptp_free_objectpropdesc	(PTPObjectPropDesc* dpd);
+void ptp_free_params		(PTPParams *params);
+
 void ptp_perror			(PTPParams* params, uint16_t error);
 
 const char*
@@ -1276,6 +1551,7 @@ int
 ptp_render_property_value(PTPParams* params, uint16_t dpc,
                           PTPDevicePropDesc *dpd, int length, char *out);
 int ptp_render_ofc(PTPParams* params, uint16_t ofc, int spaceleft, char *txt);
+int ptp_render_opcode(PTPParams* params, uint16_t opcode, int spaceleft, char *txt);
 int ptp_render_mtp_propname(uint16_t propid, int spaceleft, char *txt);
 
 /* ptpip.c */
